# -*- coding: utf-8 -*-
"""Module defining the base class for other Data types."""
from __future__ import annotations

from copy import deepcopy

from aiida import orm
from aiida.orm.nodes.data import ArrayData
import numpy as np
from phonopy import Phonopy
from phonopy.structure.cells import PhonopyAtoms

from aiida_phonopy.calculations.functions.link_structures import (
    phonopy_atoms_from_structure,
    phonopy_atoms_to_structure,
)


def _get_valid_matrix(matrix: list | np.ndarray) -> np.ndarray:
    """Get and validate the `supercell_matrix` and `primitive_matrix` inputs.

    :param matrix: (3,1) or (3,3) shape array
    :type matrix: :class:`list`, :class:`~aiida.orm.List`, :class:`numpy.ndarray`
    :return: a (3,3) numpy.ndarray
    :raises:
        * TypeError: if it is not a valid array type and if the array does not contain only numbers
        * ValueError: if the array is not of the correct shape
    """
    if not isinstance(matrix, (list, orm.List, np.ndarray)):
        raise TypeError("only `list`, `aiida.orm.List` and `numpy.ndarray` (and 'auto' as string for primitive)")

    if isinstance(matrix, np.ndarray):
        matrix = matrix.tolist()

    if not len(matrix) == 3:
        raise ValueError('need exactly (3,1) or (3,3) shape.')

    for row in matrix:
        if isinstance(row, list):
            if not len(row) in [0, 3]:
                raise ValueError('matrix need to have (3,1) or (3,3) shape.')
            for element in row:
                if not isinstance(element, (int, float)):
                    raise TypeError(
                        f'type `{type(element)}` of {element} is not an accepted '
                        'type in matrix; only `int` and `float` are valid.'
                    )

    valid_matrix = np.array(matrix)
    valid_matrix = np.diag(valid_matrix) if valid_matrix.shape == (3,) else valid_matrix

    return valid_matrix


class RawData(ArrayData):  # pylint: disable=too-many-ancestors
    """Base class containing the information for the other phonon related data types."""

    def __init__(
        self,
        structure: orm.StructureData | None = None,
        phonopy_atoms: PhonopyAtoms | None = None,
        supercell_matrix: list | None = None,
        primitive_matrix: list | None = None,
        symprec: float = 1e-05,
        is_symmetry: bool = True,
        distinguish_kinds: bool = True,
        **kwargs
    ):
        """Instantiate the class.

        The minimal input is to define either the `structure` or the `phonopy_atoms` input.
        They cannot be specified at the same time.

        :param structure: an :class:`~aiida.orm.StructureData` node
        :param phononpy_atoms: a :class:`~phonopy.structure.cells.PhonopyAtoms` instance
        :param supercell_matrix: a (3,3) shape array describing the supercell transformation
        :param primitive_matrix: a (3,3) shape array describing the primite transformation
        :param symprec: precision tollerance for symmetry analysis
        :param is_symmetry: whether using symmetries
        :distinguish_kinds: it stores a mapping between kinds and chemical symbols;
            by default Phonopy does not support kind,
            thus useful if in the input `structure` kinds are defined
        :paramm kwargs: for internal use
        """
        if structure and phonopy_atoms:
            raise ValueError('cannot pass `structure`and `phonopy_atoms` at the same time')

        if structure:
            phonopy_atoms_unitcell, mapping = phonopy_atoms_from_structure(structure)
            pbc = structure.pbc
        elif phonopy_atoms:
            phonopy_atoms_unitcell, mapping = (phonopy_atoms, None)
            pbc = [True, True, True]  # to be changed when Phonopy will support PBC
        else:
            raise ValueError('at least one between `structure` and `phonopy_atoms` must be specified')

        args = {
            'unitcell': phonopy_atoms_unitcell,
            'supercell_matrix': supercell_matrix,
            'primitive_matrix': primitive_matrix,
            'symprec': symprec,
            'is_symmetry': is_symmetry,
        }

        try:
            Phonopy(**args)
        except Exception as err:
            raise ValueError('one or more inputs are not consistent with the `phonopy.Phonopy` format') from err

        super().__init__(**kwargs)

        # Inizializing the class attributes.
        self._set_phonopy_version()
        self._set_unitcell_attributes(phonopy_atoms=phonopy_atoms_unitcell, pbc=pbc)
        self._set_kinds_map(mapping)
        self._set_distinguish_kinds(distinguish_kinds)  # crucial before setting symbols/names!
        self._set_symbols_and_names()

        if not supercell_matrix is None:
            self._set_supercell_matrix(supercell_matrix)
        else:
            self._set_supercell_matrix(np.eye(3))

        self._set_symprec(symprec)
        self._set_is_symmetry(is_symmetry)

        # Important that is the last to be specified since, if "auto" is passed,
        # a phonopy instance is generated to get the primitive structure.
        if not primitive_matrix is None:
            self._set_primitive_matrix(primitive_matrix)
        else:
            self._set_primitive_matrix('auto')

    @property
    def phonopy_version(self) -> str:
        """Get the Phonopy version used."""
        return self.base.attributes.get('phonopy_version')

    def _set_phonopy_version(self):
        """Set the installed Phonopy version."""
        from phonopy.version import __version__ as the_phonopy_version

        self._if_can_modify()
        self.base.attributes.set('phonopy_version', the_phonopy_version)

    @property
    def numbers(self) -> np.ndarray:
        """Get the array corresponding to the atomic number in periodic table of the sturcture.

        :return: (nat,) shaper array
        """
        return self.base.attributes.get('numbers')

    @property
    def masses(self) -> np.ndarray:
        """Get the array of the atomic masses in the cell.

        :return: (nat,) shape array
        """
        return self.base.attributes.get('masses')

    @property
    def positions(self) -> np.ndarray:
        """Get the array of the atomic positions in the cell.

        :return: (nat, 3) shape array
        """
        return self.base.attributes.get('positions')

    @property
    def cell(self) -> np.ndarray:
        """Get the lattice matrix of the structure.

        .. important: lattice vectors as rows of the matrix

        :return: (3,3) shape array
        """
        return self.base.attributes.get('cell')

    @property
    def magnetic_moments(self) -> list[int]:
        """Get the `magnetic_moments` array of the atoms in the cell."""
        return self.base.attributes.get('magnetic_moments')

    @property
    def symbols(self) -> list[str]:
        """Get the chemical `symbols` array of the atoms in the cell."""
        return self.base.attributes.get('symbols')

    @property
    def pbc(self) -> tuple[int, int, int]:
        """Get the periodic boundary conditions."""
        return self.base.attributes.get('pbc')

    @property
    def names(self) -> list[str]:
        """Get the custom `names` array of the atoms in the cell.

        .. note: if no special names are specified, this will be equal to `symbols`.
        """
        return self.base.attributes.get('names')

    def _set_unitcell_attributes(self, phonopy_atoms: PhonopyAtoms, pbc: tuple[bool, bool, bool]):
        """Set the attributes for full reproducibility of the `PhonopyAtoms` class."""
        self._if_can_modify()
        self.base.attributes.set_many({
            'numbers': phonopy_atoms.numbers,
            'masses': phonopy_atoms.masses,
            'positions': phonopy_atoms.positions,
            'cell': phonopy_atoms.cell,
            'magnetic_moments': phonopy_atoms.magnetic_moments,
            'pbc': pbc,
        })

    def _set_symbols_and_names(self):
        """Set the `symbols` and `names`."""
        if self.kinds_map:
            numbers_to_names, numbers_to_symbols = self.kinds_map
            symbols = []
            names = []
            for number in self.numbers:
                try:
                    names.append(numbers_to_names[number])
                    symbols.append(numbers_to_symbols[number])
                except KeyError:
                    names.append(numbers_to_names[str(number)])
                    symbols.append(numbers_to_symbols[str(number)])
            self.base.attributes.set('symbols', symbols)
            self.base.attributes.set('names', names)
        else:
            phonopy_atoms = self._get_phonopy_atoms_unitcell(distinguish_kinds=True)
            self.base.attributes.set('symbols', deepcopy(phonopy_atoms.symbols))
            self.base.attributes.set('names', deepcopy(phonopy_atoms.symbols))

    @property
    def supercell_matrix(self) -> list:
        """Get the `supercell_matrix`.

        :return: a (3,3) shape array
        """
        return self.base.attributes.get('supercell_matrix')

    def _set_supercell_matrix(self, value: list | np.ndarray):
        """Set the Phonopy supercell matrix.

        :param value: (3,3) or (3,1) shape array
        :type value: :class:`list`, :class:`~aiida.orm.List`, :class:`numpy.ndarray`

        :raises ModificationNotAllowed: if object is already stored
        """
        self._if_can_modify()
        the_supercell_matrix = _get_valid_matrix(value)

        self.base.attributes.set('supercell_matrix', deepcopy(the_supercell_matrix))

    @property
    def primitive_matrix(self) -> list:
        """Get the `primitive_matrix`.

        :return: a (3,3) shape array
        """
        return self.base.attributes.get('primitive_matrix')

    def _set_primitive_matrix(self, value: list | np.ndarray):
        """Set the primitive matrix.

        :param value: (3,3) or (3,1) shape array, or 'auto"
        :type value: :class:`list`, :class:`~aiida.orm.List`, :class:`numpy.ndarray`, :class:`str`
        :raises ModificationNotAllowed: if object is already stored
        """
        self._if_can_modify()

        if isinstance(value, str):
            the_primitive_matrix = self.get_phonopy_instance(**{'primitive': value}).primitive_matrix
        else:
            the_primitive_matrix = _get_valid_matrix(value)

        self.base.attributes.set('primitive_matrix', deepcopy(the_primitive_matrix))

    @property
    def symprec(self) -> float:
        """Get the tolerance for symmetry analysis."""
        return self.base.attributes.get('symprec')

    def _set_symprec(self, value: float):
        """Set the symmetry tolerance.

        :param value: tolerance for symmetry analysis. Check that you get
        the right symmetry of your structure before starting any calculation. Default is 1e-05.
        :type value: float
        :raises:
            * ModificationNotAllowed: if object is already stored
            * TypeError: if the input is not of type `float`
        """
        self._if_can_modify()
        if not isinstance(value, float):
            raise ValueError('only `float` type is accepted.')

        self.base.attributes.set('symprec', deepcopy(value))

    @property
    def is_symmetry(self) -> bool:
        """Get `is_symmetry` value.

        It refers to whether Phonopy will use symmetries to reduce the number of
        displacements for frozen phonons.
        """
        return self.base.attributes.get('is_symmetry')

    def _set_is_symmetry(self, value: bool):
        """Set whether to use the symmetries.

        Use with care and if you know what your are doing.

        :param value: whether to use or not the symmetries. Deafault is True.
        :type value: bool

        :raises:
            * ModificationNotAllowed: if object is already stored
            * TypeError: if the input is not of type `bool`
        """
        self._if_can_modify()
        if not isinstance(value, bool):
            raise ValueError('only `bool` type is accepted.')

        self.base.attributes.set('is_symmetry', deepcopy(value))

    @property
    def kinds_map(self) -> dict | None:
        """Get the map bewtween of the `numbers` and the `symbols` and `names`."""
        try:
            the_map = self.base.attributes.get('kinds_map')
        except AttributeError:
            the_map = None
        return the_map

    def _set_kinds_map(self, value: dict):
        """Set the kind names map between the PhonopyAtoms unitcell and a reference structure.

        This is needed since PhonopyAtoms does not support kind names.
        This attribute allows to get proper `StructureData` supercells with displacements.

        :param value: tuple with two dictionaries (numbers_to_names, numbers_to_symbols)
        """
        self._if_can_modify()
        self.base.attributes.set('kinds_map', value)

    @property
    def distinguish_kinds(self) -> bool:
        """Get whether or not kinds with same chemical symbol will be distinguished by symmetry."""
        return self.base.attributes.get('distinguish_kinds')

    def _set_distinguish_kinds(self, value: bool):
        """Set whether or not to distinguish kinds."""
        self._if_can_modify()
        if not isinstance(value, bool):
            raise ValueError('only `bool` type is accepted.')

        self.base.attributes.set('distinguish_kinds', value)

    def _get_phonopy_atoms_unitcell(self, distinguish_kinds: bool) -> PhonopyAtoms:
        """Get the PhonopyAtoms object using the internal attributes."""
        kwargs = {
            'cell': self.cell,
            'positions': self.positions,
            'masses': self.masses,
            'magmoms': self.magnetic_moments
        }
        if distinguish_kinds:
            kwargs.update({'numbers': self.numbers})
        else:
            kwargs.update({'symbols': self.symbols})

        return PhonopyAtoms(**kwargs)

    def get_phonopy_instance(
        self, symmetrize_nac: bool | None = None, factor_nac: float | None = None, **kwargs
    ) -> Phonopy:
        """Return a :py:class:`phonopy.Phonopy` object with the current values.

        :param symmetrize_nac: whether or not to symmetrize the nac parameters using point group symmetry;
            defaults to self.is_symmetry
        :type symmetrize_nac: bool
        :param factor_nac: factor for non-analytical corrections; defaults to Hartree*Bohr
        :type factor_nac: float
        :param kwargs: for internal use to set the primitive cell
        """
        from phonopy.structure.symmetry import symmetrize_borns_and_epsilon
        from phonopy.units import Bohr, Hartree

        distinguish = kwargs.pop('distinguish', self.distinguish_kinds)
        unitcell = self._get_phonopy_atoms_unitcell(distinguish)

        primitive_matrix = kwargs.pop('primitive', None)
        if primitive_matrix is None:
            primitive_matrix = self.primitive_matrix

        ph_instance = Phonopy(
            unitcell=unitcell,
            supercell_matrix=self.supercell_matrix,
            primitive_matrix=primitive_matrix,
            symprec=self.symprec,
            is_symmetry=self.is_symmetry
        )

        # Non-analytical parameters
        if not isinstance(symmetrize_nac, bool) and symmetrize_nac is not None:
            raise TypeError('`symmetrize_nac` not of the right type')

        if symmetrize_nac is not None:
            do_symmetrize = symmetrize_nac
        else:
            do_symmetrize = self.is_symmetry

        if not isinstance(factor_nac, (float, int)) and factor_nac is not None:
            raise TypeError('input not of the right type')

        if factor_nac is not None:
            the_factor = factor_nac
        else:
            the_factor = Hartree * Bohr

        if self.has_nac_parameters():
            if do_symmetrize:
                nac = symmetrize_borns_and_epsilon(
                    borns=self.born_charges, epsilon=self.dielectric, ucell=ph_instance.primitive, symprec=self.symprec
                )
                the_born = nac[0]
                the_dielectric = nac[1]
            else:
                the_born = self.born_charges
                the_dielectric = self.dielectric

            ph_instance.nac_params = {'born': the_born, 'dielectric': the_dielectric, 'factor': the_factor}

        return ph_instance

    def get_unitcell(self) -> orm.StructureData:
        """Get the `unitcell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).unitcell
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map, self.pbc)

    def get_primitive_cell(self) -> orm.StructureData:
        """Get the `primitive cell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).primitive
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map, self.pbc)

    def get_supercell(self) -> orm.StructureData:
        """Get the pristine `supercell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).supercell
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map, self.pbc)

    def get_cells_mappings(self) -> dict[dict[list]]:
        """Return a dictionary containing the mappings among unit-, super- and primitive cell.

        :return: dictionary with the following key:pair structure:
            * primitive: {p2p_map: list, p2s_map: list, s2p_map: list}
            * supercell: {u2u_map: list, u2s_map: list, s2u_map: list}
        """
        ph = self.get_phonopy_instance()

        cells_maps = {
            'primitive': {
                'p2p_map': ph.primitive.p2p_map,
                'p2s_map': ph.primitive.p2s_map,
                's2p_map': ph.primitive.s2p_map,
            },
            'supercell': {
                'u2u_map': ph.supercell.u2u_map,
                'u2s_map': ph.supercell.u2s_map,
                's2u_map': ph.supercell.s2u_map,
            },
        }

        return cells_maps

    @property
    def dielectric(self) -> np.ndarray | None:
        """Get the high-frequency dielectric tensor in Cartesian coordinates."""
        try:
            value = self.base.attributes.get('dielectric')
            value = np.array(value)
        except (KeyError, AttributeError):
            value = None
        return value

    def set_dielectric(self, dielectric: list | np.ndarray):
        """Set the high-frequency dielectric tensor in Cartesian coordinates.

        .. note: it is assumed that the reference system is the same of the primitive cell.

        :param dielectric: (3, 3) array like

        :raises:
            * TypeError: if the format is not compatible or of the correct type
            * ValueError: if the format is not compatible or of the correct type
        """
        self._if_can_modify()

        if not isinstance(dielectric, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        the_dielectric = np.array(dielectric)

        if the_dielectric.shape == (3, 3):
            self.base.attributes.set('dielectric', the_dielectric.tolist())
        else:
            raise ValueError('the array is not of the correct shape')

    @property
    def born_charges(self) -> np.ndarray:
        """Get the effective Born charges tensors in Cartesian coordinates.

        ..note:
            The indecis refers to:
                1. Atomic index.
                2. Polarization index.
                3. Atomic displacement index.

        :returns: numpy.ndarray, shape (num primitive cell atoms, 3, 3)
        """
        try:
            value = self.get_array('born_charges')
        except (KeyError, AttributeError):
            value = None
        return value

    def set_born_charges(self, born_charges: list | np.ndarray):
        """Set the Born effective charge tensors in Cartesian coordinates.

        ..note:
            The indecis refers to:
                1. Atomic index.
                2. Polarization index.
                3. Atomic displacement index.

        :param born_charges: (number of atoms in the primitive cell, 3, 3) shape array like

        :raises:
            * TypeError: if the format is not compatible or of the correct type
            * ValueError: if the format is not compatible or of the correct type
        """
        self._if_can_modify()

        if not isinstance(born_charges, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        the_born_charges = np.array(born_charges)

        nprimitive_atoms = len(self.get_primitive_cell().sites)

        if the_born_charges.shape == (nprimitive_atoms, 3, 3):
            self.set_array('born_charges', the_born_charges)
        else:
            raise ValueError('the array is not of the correct shape')

    def has_nac_parameters(self) -> bool:
        """Return wheter or not the Data has non-analytical constants."""
        return (self.dielectric is not None and self.born_charges is not None)

    def _if_can_modify(self):
        """Check if the object is stored and raise an error if so. To use in every setter."""
        from aiida.common.exceptions import ModificationNotAllowed

        if self.is_stored:
            raise ModificationNotAllowed('The PhonopyData object cannot be modified, it has already been stored')



# -*- coding: utf-8 -*-
"""Module defining the class for managing the frozen phonon structure."""
from __future__ import annotations

import copy
import json

from aiida import orm
import numpy as np
from phonopy.structure.cells import PhonopyAtoms

from aiida_phonopy.calculations.functions.link_structures import phonopy_atoms_to_structure

from .raw import RawData


class PreProcessData(RawData):  # pylint: disable=too-many-ancestors
    """Class for pre-processing of frozen-phonon calculations.

    This class is designed for handling the pre-process information regarding
    a frozen phonon calculation using ``Phonopy``. These regard the unitcell structure,
    the supercell and primitive matrix, as well as other symmetry information.
    """

    def __init__(
        self,
        structure: orm.StructureData | None = None,
        phonopy_atoms: PhonopyAtoms | None = None,
        supercell_matrix: list | None = None,
        primitive_matrix: list | None = None,
        symprec: float = 1e-05,
        is_symmetry: bool = True,
        distinguish_kinds: bool = True,
        **kwargs
    ):
        """Instantiate the class.

        The minimal input is to define either the `structure` or the `phonopy_atoms` input.
        They cannot be specified at the same time.

        :param structure: an :class:`~aiida.orm.StructureData` node
        :param phononpy_atoms: a :class:`~phonopy.structure.cells.PhonopyAtoms` instance
        :param supercell_matrix: a (3,3) shape array describing the supercell transformation
        :param primitive_matrix: a (3,3) shape array describing the primite transformation
        :param symprec: precision tollerance for symmetry analysis
        :param is_symmetry: whether using symmetries
        :distinguish_kinds: it stores a mapping between kinds and chemical symbols;
            by default Phonopy does not support kind,
            thus useful if in the input `structure` kinds are defined
        :paramm kwargs: for internal use
        """
        kwargs['structure'] = structure
        kwargs['phonopy_atoms'] = phonopy_atoms
        kwargs['supercell_matrix'] = supercell_matrix
        kwargs['primitive_matrix'] = primitive_matrix
        kwargs['symprec'] = symprec
        kwargs['is_symmetry'] = is_symmetry
        kwargs['distinguish_kinds'] = distinguish_kinds

        super().__init__(**kwargs)

        if self.__class__.__name__ == 'PreProcessData':
            self.set_displacements()

    @property
    def displacement_dataset(self) -> dict | list | None:
        """Get the dispacement dataset in a readible format for phonopy.

        If not set, None is returned.
        """
        message = '`displacement_dataset` stored in the database will be deprecated in v2.0.0'
        try:
            import warnings
            the_dataset = self.base.attributes.get('displacement_dataset')
            warnings.warn(message, DeprecationWarning)
            return the_dataset
        except AttributeError:
            filename = 'displacement_dataset.json'

            if filename in self.base.repository.list_object_names():
                with self.base.repository.open(filename, mode='rb') as handle:
                    return json.load(handle)

            return None

    @property
    def displacements(self) -> list | dict:
        """Get the displacements to apply to the supercell.

        :returns: array with displacements; can be type-I or type-II (see :func:`phonopy.Phonopy.displacements`)
        """
        # For the time being, it is important to keep this implementation for the subclasses in the package,
        # otherwise a loop is generated (i.e. in the PhonopyData class, when setting the forces).
        ph = super().get_phonopy_instance()
        ph.dataset = self.displacement_dataset
        return ph.displacements

    def get_displacements(self) -> list | dict:
        """Get the displacements to apply to the supercell."""
        return self.displacements

    def set_displacements(
        self,
        distance: float = 0.01,
        is_plusminus: str = 'auto',
        is_diagonal: bool = True,
        is_trigonal: bool = False,
        number_of_snapshots: int | None = None,
        random_seed: int | None = None,
        temperature: float | None = None,
        cutoff_frequency: float | None = None,
    ):
        """Set displacements for frozen phonon calculation.

        Refer to :py:func:`phonopy.Phonopy.generate_displacements` for a complete description of the inputs.

        .. note: temperature different from 0K can be given only when forces are set up.
        Thus, in the PreProcessData value different from zero will raise an error.

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        self._if_can_modify()
        ph = self.get_phonopy_instance()

        try:
            ph.generate_displacements(
                distance=distance,
                is_plusminus=is_plusminus,
                is_diagonal=is_diagonal,
                is_trigonal=is_trigonal,
                number_of_snapshots=number_of_snapshots,
                random_seed=random_seed,
                temperature=temperature,
                cutoff_frequency=cutoff_frequency,
            )
        except ValueError as err:
            raise ValueError('one or more input types are not accepted') from err

        self._set_displacements(copy.deepcopy(ph.dataset))

    def _set_displacements(self, value: list | dict):
        """Put in the repository the displacement dataset in json format."""
        try:
            serialized = json.dumps(value)
        except TypeError:
            serialized = json.dumps(_serialize(value))

        self.base.repository.put_object_from_bytes(serialized.encode('utf-8'), 'displacement_dataset.json')

    def set_displacements_from_dataset(self, dataset: dict | list):
        """Set displacements for frozen phonon calculation from a dataset.

        Useful if you want to set displacements from a previously random generated
        displacement dataset, or for setting dataset for self-consistent harmonic
        approximation.

        :param dataset: dictionary or array like (numpy or list), compatible with phonopy

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        self._if_can_modify()
        ph = self.get_phonopy_instance()

        if isinstance(dataset, dict):
            try:
                ph.dataset = dataset
            except RuntimeError as err:
                raise ValueError('non compatible format') from err
        elif isinstance(dataset, (np.ndarray, list)):
            try:
                ph.displacements = dataset
            except RuntimeError as err:
                raise ValueError('non compatible format') from err
        else:
            raise ValueError('type not accepted')

        self._set_displacements(_serialize(copy.deepcopy(ph.dataset)))

    def get_phonopy_instance(self, symmetrize_nac: bool = None, factor_nac: float | None = None, **kwargs):
        """Return a :class:`~phonopy.Phonopy` object with the current values.

        :param symmetrize_nac: whether or not to symmetrize the nac parameters using point group symmetry;
            defaults to self.is_symmetry
        :type symmetrize_nac: bool
        :param factor_nac: factor for non-analytical corrections; defaults to Hartree*Bohr
        :type factor_nac: float
        :param kwargs: for internal use to set the primitive cell
        """
        ph = super().get_phonopy_instance(symmetrize_nac, factor_nac, **kwargs)
        if self.displacement_dataset is not None:
            ph.dataset = self.displacement_dataset
        return ph

    def get_supercells_with_displacements(self) -> dict[orm.StructureData]:
        """Get the supercells with displacements for frozen phonon calculation.

        .. note: this is not linking in the provenance the output structures.
            Use the `self.calcfunctions.get_supercells_with_displacements` instead

        :returns: dictionary with StructureData nodes, None if
            the displacement dataset has not been set
        """
        # Here we distinguish the kind, but only for mapping purposes.
        # This does not affect the number of displacements, since they
        # have been set with the correct flag of not distinguishing kinds.
        ph = self.get_phonopy_instance(**{'distinguish': True})

        supercells = ph.supercells_with_displacements
        mapping = self.kinds_map
        structures_dict = {}

        digits = len(str(len(supercells)))  # this will make the labels more nice to read

        for i, scell in enumerate(supercells):  # start from 1 - better choice for gathering forces
            label = f'supercell_{str(i + 1).zfill(digits)}'
            structures_dict.update({label: phonopy_atoms_to_structure(scell, mapping, self.pbc)})

        return structures_dict

    def generate_displacement_dataset(
        self,
        distance: float = 0.01,
        is_plusminus: str = 'auto',
        is_diagonal: bool = True,
        is_trigonal: bool = False,
        number_of_snapshots: int | None = None,
        random_seed: int | None = None,
        temperature: float | None = None,
        cutoff_frequency: float | None = None,
    ):
        """Return the displacement dataset for frozen phonon calculation.

        Refer to :py:func:`phonopy.Phonopy.generate_displacements` for a complete description of the inputs.

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        ph = self.get_phonopy_instance()

        try:
            ph.generate_displacements(
                distance=distance,
                is_plusminus=is_plusminus,
                is_diagonal=is_diagonal,
                is_trigonal=is_trigonal,
                number_of_snapshots=number_of_snapshots,
                random_seed=random_seed,
                temperature=temperature,
                cutoff_frequency=cutoff_frequency,
            )
        except ValueError as err:
            raise ValueError('one or more inputs made the phonopy instance to raise an error') from err

        return copy.deepcopy(ph.dataset)

    @property
    def calcfunctions(self):
        """Namespace to access the calcfunction utilities."""
        from aiida_phonopy.calculations.functions.data_utils import CalcfunctionMixin

        return CalcfunctionMixin(data_node=self)

    @staticmethod
    def generate_preprocess_data(
        structure: orm.StructureData,
        displacement_generator: dict | None = None,
        supercell_matrix: list | None = None,
        primitive_matrix: list | None = None,
        symprec: float | None = None,
        is_symmetry: bool | None = None,
        distinguish_kinds: bool | None = None,
    ):
        """Return a complete stored PreProcessData node.

        :param structure: structure data node representing the unitcell
        :type structure: :class:`~aiida.orm.StructureData`
        :param displacement_generator: dictionary containing the info for generating the displacements,
            defaults to phonopy default (see phonopy doc)
        :type displacement_generator: orm.Dict
        :param supercell_matrix: supercell matrix, defaults to diag(1,1,1)
        :type supercell_matrix: :class:`~aiida.orm.List`, Optional
        :param primitive_matrix: primitive matrix, defaults to "auto"
        :type primitive_matrix: :class:`~aiida.orm.List`, Optional
        :param symprec: symmetry precision on atoms, defaults to 1e-5
        :type symprec: :class:`~aiida.orm.Float`, Optional
        :param is_symmetry: if using space group symmetry, defaults to True
        :type is_symmetry: :class:`~aiida.orm.Bool`, Optional
        :param distinguish_kinds: if distinguish names of same specie by symmetry, defaults to True
        :type distinguish_kinds: :class:`~aiida.orm.Bool`, Optional

        :return: PreProcessData node
        """
        from aiida_phonopy.calculations.functions.data_utils import generate_preprocess_data

        kwargs = {}

        kwargs['structure'] = structure

        if displacement_generator is not None:
            kwargs['displacement_generator'] = displacement_generator

        if supercell_matrix is not None:
            kwargs['supercell_matrix'] = supercell_matrix

        if primitive_matrix is not None:
            kwargs['primitive_matrix'] = primitive_matrix

        if symprec is not None:
            kwargs['symprec'] = symprec

        if is_symmetry is not None:
            kwargs['is_symmetry'] = is_symmetry

        if distinguish_kinds is not None:
            kwargs['distinguish_kinds'] = distinguish_kinds

        return generate_preprocess_data(**kwargs)


def _serialize(data: dict | list):
    """Serialize the data for displacement dataset, in case it contains numpy.ndarray."""
    serialized = copy.deepcopy(data)

    try:
        for key, value in data.items():
            serialized[key] = _serialize(value)
    except AttributeError:
        if isinstance(serialized, list):
            for i, value in enumerate(data):
                serialized[i] = _serialize(value)
        else:
            if isinstance(data, np.ndarray):
                return copy.deepcopy(data.tolist())

    return serialized



# -*- coding: utf-8 -*-
"""Module defining the class for force constants data."""
from __future__ import annotations

from aiida.orm import StructureData
import numpy as np
from phonopy.structure.cells import PhonopyAtoms

from .raw import RawData


class ForceConstantsData(RawData):  # pylint: disable=too-many-ancestors
    """Self-contained class for force constants data and non-analytical constants.

    It stores also the structure information (unitcell, supercell, ...), for a
    complete transferable data type.
    """

    def __init__(
        self,
        structure: StructureData | None = None,
        phonopy_atoms: PhonopyAtoms | None = None,
        supercell_matrix: list | None = None,
        primitive_matrix: list | None = None,
        symprec: float = 1e-05,
        is_symmetry: bool = True,
        distinguish_kinds: bool = True,
        **kwargs
    ):
        """Instantiate the class.

        The minimal input is to define either the `structure` or the `phonopy_atoms` input.
        They cannot be specified at the same time.

        :param structure: a :class:`~aiida.orm.StructureData` node
        :param phononpy_atoms: a :class:`~phonopy.structure.cells.PhonopyAtoms` instance
        :param supercell_matrix: a (3,3) shape array describing the supercell transformation
        :param primitive_matrix: a (3,3) shape array describing the primite transformation
        :param symprec: precision tollerance for symmetry analysis
        :param is_symmetry: whether using symmetries
        :distinguish_kinds: it stores a mapping between kinds and chemical symbols;
            by default Phonopy does not support kind,
            thus useful if in the input `structure` kinds are defined
        :paramm kwargs: for internal use
        """
        kwargs['structure'] = structure
        kwargs['phonopy_atoms'] = phonopy_atoms
        kwargs['supercell_matrix'] = supercell_matrix
        kwargs['primitive_matrix'] = primitive_matrix
        kwargs['symprec'] = symprec
        kwargs['is_symmetry'] = is_symmetry
        kwargs['distinguish_kinds'] = distinguish_kinds

        super().__init__(**kwargs)

    def get_phonopy_instance(self, **kwargs):
        """Return a :class:`~phonopy.Phonopy` object with force and nac parameters (if set).

        :param kwargs: see :func:`aiida_phonopy.data.preprocess.PreProcessData.get_phonopy_instance`
            * symmetrize_nac: whether or not to symmetrize the nac parameters
                using point group symmetry; bool, defaults to self.is_symmetry
            * factor_nac: factor for non-analytical corrections;
                float, defaults to Hartree*Bohr
        """
        ph_instance = super().get_phonopy_instance(**kwargs)

        if self.force_constants is not None:
            ph_instance.force_constants = self.force_constants

        return ph_instance

    @property
    def force_constants(self) -> np.ndarray:
        """Get force force constants matrix."""
        try:
            the_forces = self.get_array('force_constants')
        except KeyError:
            the_forces = None
        return the_forces

    def set_force_constants(self, force_constants: list | np.ndarray):
        """Set force constants matrix.

        :param force_constants: array of force constants matrix in compact or full format
        :param type: (n_patom = atoms in primitive cell, n_satom =  atoms in supercell)
            * Compact format: (n_patom, n_satom, 3, 3)
            * Full format: (n_satom, n_satom, 3, 3)

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
            * RuntimeError: if the displacement dataset was not initialize in input
        """
        self._if_can_modify()

        if not isinstance(force_constants, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        n_satoms = len(self.get_supercell().sites)
        n_patoms = len(self.get_primitive_cell().sites)

        fc = np.array(force_constants)

        if fc.shape in [(n_patoms, n_satoms, 3, 3), (n_satoms, n_satoms, 3, 3)]:
            self.set_array('force_constants', fc)
        else:
            raise ValueError('the array is not of the correct shape')



# -*- coding: utf-8 -*-
"""Validate that the version in the tag label matches the version of the package."""
import argparse
import ast
from pathlib import Path


def get_version_from_module(content: str) -> str:
    """Get the ``__version__`` attribute from a module.

    .. note:: This has been adapted from :mod:`setuptools.config`.
    """
    try:
        module = ast.parse(content)
    except SyntaxError as exception:
        raise IOError('Unable to parse module.') from exception

    try:
        return next(
            ast.literal_eval(statement.value) for statement in module.body if isinstance(statement, ast.Assign)
            for target in statement.targets if isinstance(target, ast.Name) and target.id == '__version__'
        )
    except StopIteration as exception:
        raise IOError('Unable to find the `__version__` attribute in the module.') from exception


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('GITHUB_REF', help='The GITHUB_REF environmental variable')
    args = parser.parse_args()
    assert args.GITHUB_REF.startswith('refs/tags/v'), f'GITHUB_REF should start with "refs/tags/v": {args.GITHUB_REF}'
    tag_version = args.GITHUB_REF[11:]
    package_version = get_version_from_module(Path('src/aiida_phonopy/__init__.py').read_text(encoding='utf-8'))
    error_message = f'The tag version `{tag_version}` is different from the package version `{package_version}`'
    assert tag_version == package_version, error_message



# -*- coding: utf-8 -*-
"""Module defining the class which wraps the :class:`phonopy.Phonopy` main class."""
from __future__ import annotations

import numpy as np
from phonopy import Phonopy

from .preprocess import PreProcessData


class PhonopyData(PreProcessData):  # pylint: disable=too-many-ancestors
    """Class wrapping the :class:`phonopy.Phonopy` class.

    It represents the final Data node status of a frozen phonon calculaiton.
    It stores information regarding the pre-processing, the displacements and
    forces dataset, and the (eventual) non-analytical constants.

    .. note: direct calculation of properties from this class is still not implemented.
        Use :class:`~aiida_phonopy.calculations.phonopy.PhonopyCalculation` for
        post-processing this data node, keeping the provenance and having the
        data produced in the correct data type.
    """

    def __init__(self, preprocess_data: PreProcessData, **kwargs):
        """Instantiate the class.

        :param preprocess_data: a :class:`~aiida_phonopy.data.preprocess.PreProcessData` node

        :raises TypeError: if the input is not of the correct type,
        :raises ValueError: if the preprocess_data input does not contain displacement dataset
        """
        if not isinstance(preprocess_data, PreProcessData):
            raise TypeError(f'incorrect type. Given type <{type(preprocess_data)}>, expected `PreProcessData')

        kwargs['structure'] = preprocess_data.get_unitcell()
        kwargs['supercell_matrix'] = preprocess_data.supercell_matrix
        kwargs['primitive_matrix'] = preprocess_data.primitive_matrix
        kwargs['symprec'] = preprocess_data.symprec
        kwargs['is_symmetry'] = preprocess_data.is_symmetry
        kwargs['distinguish_kinds'] = preprocess_data.distinguish_kinds
        super().__init__(**kwargs)

        dataset = preprocess_data.displacement_dataset

        if dataset is not None:
            super().set_displacements_from_dataset(dataset)
        else:
            raise ValueError('cannot instantiate object without having displacement dataset set')

    def set_displacements(self):
        """Set displacements cannot be accessed from PhonopyData."""
        raise RuntimeError('`displacements` cannot be changed for this Data')

    def set_displacements_from_dataset(self):
        """Set displacements cannot be accessed from PhonopyData."""
        raise RuntimeError('`displacements` cannot be changed for this Data')

    def get_phonopy_instance(self, subtract_residual_forces: bool | None = None, **kwargs) -> Phonopy:
        """Return a :class:`~phonopy.Phonopy` object with forces and nac parameters (if set).

        :param subtract_residual_forces: whether or not subract residual forces (if set)
        :type subtract_residual_forces: bool
        :param kwargs: see :func:`aiida_phonopy.data.preprocess.PreProcessData.get_phonopy_instance`
            * symmetrize_nac: whether or not to symmetrize the nac parameters
                using point group symmetry; bool, defaults to self.is_symmetry
            * factor_nac: factor for non-analytical corrections;
                float, defaults to Hartree*Bohr
        """
        if not isinstance(subtract_residual_forces, bool) and subtract_residual_forces is not None:
            raise TypeError('`subtract_residual_forces` not of the right type')

        if subtract_residual_forces and self.residual_forces is None:
            raise ValueError('`subtract_residual_forces` is set to True, but `residual_forces` are None')

        ph_instance = super().get_phonopy_instance(**kwargs)

        if self.displacement_dataset is not None:
            ph_instance.dataset = self.displacement_dataset

        try:
            if self.forces is not None:
                the_forces = self.forces
                if subtract_residual_forces:
                    the_forces = the_forces - self.residual_forces
                ph_instance.forces = the_forces
        except AttributeError:
            pass

        return ph_instance

    @property
    def residual_forces(self) -> np.ndarray:
        """Get the residual forces calculated on the pristine (i.e. no displaced) supercell structure (if set).

        ..note: if you have specified the `forces_index` this will be used as well here.
        """
        try:
            if self.forces_index is not None:
                the_forces = self.get_array('residual_forces')[self.forces_index]
            else:
                the_forces = self.get_array('residual_forces')
        except KeyError:
            the_forces = None
        return the_forces

    def set_residual_forces(self, forces: list | np.ndarray):
        """Set the residual forces of the pristine supercell.

        :param forces: (atoms in supercell, 3) array shape

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
        """
        self._if_can_modify()

        if not isinstance(forces, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        natoms = len(self.get_supercell().sites)

        the_forces = np.array(forces)

        if self.forces_index is not None:
            if the_forces[self.forces_index].shape == (natoms, 3):
                self.set_array('residual_forces', the_forces)
            else:
                raise ValueError('the array is not of the correct shape')
        else:
            if the_forces.shape == (natoms, 3):
                self.set_array('residual_forces', the_forces)
            else:
                raise ValueError('the array is not of the correct shape. Check also `forces_index`')

    @property
    def forces(self) -> np.ndarray:
        """Get forces for each supercell with displacements in the dataset as a unique array."""
        try:
            the_forces = self.get_array('forces')
            return the_forces
        except (KeyError, AttributeError):
            pass

        if not 'forces_1' in self.get_arraynames():
            return

        try:
            nsupercells = len(self.displacements)
            the_forces = np.zeros((nsupercells)).tolist()
            for i in range(nsupercells):
                if self.forces_index is not None:
                    the_forces[i] = self.get_array(f'forces_{i+1}')[self.forces_index]
                else:
                    the_forces[i] = self.get_array(f'forces_{i+1}')
            the_forces = np.array(the_forces)
        except (KeyError, AttributeError):
            return

        return the_forces

    def set_forces(
        self,
        sets_of_forces: list | np.ndarray | None = None,
        dict_of_forces: dict | None = None,
        forces_index: int | None = None
    ):
        """Set forces per each supercell with displacement in the dataset.

        :param sets_of_forces: a set of atomic forces in displaced supercells. The order of
            displaced supercells has to match with that in displacement dataset.
        :param type: (supercells with displacements, atoms in supercell, 3) array shape
        :param dict_of_forces: dictionary of forces, in numpy.ndarray to store for each displacement.
            They keys for the dictionary must be passed as `forces_{num}`, where `num` corresponds to
            the associated supercell in the dataset. `num` starts from 1.
        :param forces_index: an integer storing in the database the index for forces. The `dict_of_forces`
            may be specified from `TrajectoryData` to reduce the amount of data saved in the repository.
            For example: forces_1 = [[actual array]] ==> forces_index = 0

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
            * RuntimeError: if the displacement dataset was not initialize in input
        """
        self._if_can_modify()

        if self.displacement_dataset is None:
            raise RuntimeError('the displacement dataset has not been set yet')

        nsupercells = len(self.displacements)
        natoms = len(self.get_supercell().sites)

        if sets_of_forces is not None:
            the_forces = np.array(sets_of_forces)

            if the_forces.shape == (nsupercells, natoms, 3):
                self.set_array('forces', the_forces)
            else:
                raise ValueError('the array is not of the correct shape')

        if dict_of_forces is not None:
            # First, verify the number of keys is correct
            if len(list(dict_of_forces.keys())) != nsupercells:
                raise ValueError('the dictionary does not have the correct number of forces')
            # Second, verify the keys
            for key in dict_of_forces.keys():
                if key.split('_')[0] != 'forces':
                    raise ValueError(f'{key} is not correct. Expected `forces_num` as key')
            # Third, store
            for key, value in dict_of_forces.items():
                new_key = f"forces_{int(key.split('_')[-1])}"
                self.set_array(new_key, np.array(value))

        if forces_index is not None:
            if isinstance(forces_index, int):
                self.base.attributes.set('forces_index', forces_index)
            else:
                raise ValueError('index for forces must be an integer')

    @property
    def forces_index(self) -> int:
        """Return the index of the forces to use."""
        try:
            index = self.base.attributes.get('forces_index')
        except (KeyError, AttributeError):
            index = None
        return index

    def set_forces_index(self, value: int):
        """Set the `forces_index` attribute.

        This is used for tacking a particular array index of each single forces set.
        This is useful to not duplicate data of e.g. trajectories.
        """
        if isinstance(value, int):
            self.base.attributes.set('forces_index', value)
        else:
            raise ValueError('index for forces must be an integer')
